﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Net.Sockets;
using System.Diagnostics;
using System.Net;
using ThinkGeo.GpstrackingServer.Geteways.GatewaySdk;

namespace ThinkGeo.GpstrackingServer.Geteways.TcpSampleListener
{
    public partial class TcpSampleListener : Form
    {
        // Sample Message: "351277001110161$GPRMC,155726,A,3857.2284,N,09516.9031,W,000.0,208.3,170106,002.4,E,A*18,AUTO";

        // Here is your Api key
        private const string ApiKey = "<Your Api Key>";
        // Here is your gateway web service
        private const string GatewayWebServiceUri = "https://gpsgateway.thinkgeo.com";
        // Here is the Udp port
        private const int Port = 1234;

        private GatewayManager mGatewayManager;
        private Socket mListener = null;

        public TcpSampleListener()
        {
            InitializeComponent();
        }

        private void btnStart_Click(object sender, EventArgs e)
        {
            try
            {
                mGatewayManager = new GatewayManager(new Uri(GatewayWebServiceUri), ApiKey);
                mListener = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
                IPEndPoint EndPoint = new IPEndPoint(IPAddress.Any, Port);
                mListener.Bind(EndPoint);
                mListener.Listen(1000);
                mListener.BeginAccept(new AsyncCallback(this.OnConnectRequest), mListener);
            }
            catch (Exception Ex)
            {
                System.Diagnostics.Trace.WriteLine("Maybe you need specific your first local IP address," + Ex.ToString());
            }
        }

        /// <summary>
        /// Callback used when a client requests a connection. 
        /// Accpet the connection, adding it to our list and setup to 
        /// accept more connections.
        /// </summary>
        /// <param name="ar"></param>
        public void OnConnectRequest(IAsyncResult ar)
        {
            Socket listener = (Socket)ar.AsyncState;
            try
            {
             
                NewConnection(listener.EndAccept(ar));

                GC.Collect();
                GC.WaitForPendingFinalizers();
            }
            catch (Exception ex)
            {
                System.Diagnostics.Trace.WriteLine(ex.ToString());
            }
            finally
            {
                listener.BeginAccept(new AsyncCallback(OnConnectRequest), listener);
            }
        }

        /// <summary>
        /// Add the given connection to our list of clients
        /// Setup a callback to recieve data
        /// </summary>
        /// <param name="sockClient">Connection to keep</param>
        public void NewConnection(Socket sockClient)
        {
            // Program blocks on Accept() until a client connects.			
            TcpSocketClient Client = new TcpSocketClient(sockClient);

            System.Diagnostics.Trace.WriteLine(string.Format("Device {0} connected", Client.Sock.RemoteEndPoint));
            Client.DeviceId = "";
            Client.SetupRecieveCallback(this);
            GC.Collect();
        }

        /// <summary>
        /// Get the new data
        /// The message format is like: 
        /// Device ID | General Option | Sequence No | Packet Length | Event ID | Payload Mask | Len | RF Info | Len | Dig. io | Power | Analog Input 1 | Analog Input 2 | Analog Input 3 | Len | GPS (NMEA-RMC)
        /// 8             4                  4                  2                  2           4        1       4       1         1           4           4                         4       4    1       67
        /// </summary>
        /// <param name="ar"></param>
        public void OnRecievedData(IAsyncResult ar)
        {
            TcpSocketClient Client = null;
            try
            {
                Client = (TcpSocketClient)ar.AsyncState;

                byte[] byteMsg = Client.GetRecievedData(ar);
                string rawMessage = Encoding.ASCII.GetString(byteMsg);
                ProcessRecvMsgs(rawMessage);
            }
            catch (Exception Ex)
            {
                System.Diagnostics.Trace.WriteLine("OnRecievedData Caught Error - " + Ex.Message);
            }
            finally
            {
                FinallyRecieved(Client);
            }
        }

        /// <summary>
        /// Main logic here for handle received message
        /// </summary>
        /// <param name="receiveString">The raw message</param>
        private void ProcessRecvMsgs(string receiveString)
        {
            string DeviceId = "";
            try
            {
                DeviceId = receiveString.Substring(0, 15);
                DateTime UtcDateTime = DateTime.MinValue;
                int AltitudeInMeters = 0;
                string RawDebugMessage = receiveString;
                receiveString = receiveString.Replace(DeviceId, "");
                receiveString = receiveString.Replace("$", "");
                string[] Parts = receiveString.Split(Convert.ToChar(","));
                string ValidRecord = Parts[2].ToString();
                int GpsAccuracyInMeters = -1;
                if (ValidRecord == "A")
                {
                    GpsAccuracyInMeters = 1;
                }

                string TimeString = Parts[1].ToString();
                int Hour, Minute, Second;
                Hour = Convert.ToInt32(TimeString.Substring(0, 2));
                Minute = Convert.ToInt32(TimeString.Substring(2, 2));
                Second = Convert.ToInt32(TimeString.Substring(4, 2));
                string DateString = Parts[9].ToString();

                int Day, Month, Year;
                Day = Convert.ToInt32(DateString.Substring(0, 2));
                Month = Convert.ToInt32(DateString.Substring(2, 2));
                Year = Convert.ToInt32(DateString.Substring(4, 2)) + 2000;

                UtcDateTime = new DateTime(Year, Month, Day, Hour, Minute, Second);
                string LatitudeString = Parts[3].ToString();
                string LatitudeDirection = Parts[4].ToString();
                decimal Latitude = DMToDecimalDegrees(Convert.ToDouble(LatitudeString.Substring(0, 2)), Convert.ToDouble(LatitudeString.Substring(2, 7)), LatitudeDirection);

                string LongitudeString = Parts[5].ToString();
                string LongitudeDirection = Parts[6].ToString();
                decimal Longitude = DMToDecimalDegrees(Convert.ToDouble(LongitudeString.Substring(0, 3)), Convert.ToDouble(LongitudeString.Substring(3, 7)), LongitudeDirection);

                int Speed = Convert.ToInt32(Convert.ToDouble(Parts[7].Substring(0, 3)) * 1.852);
                int SpeedInKph = (int)(Speed * 1.85200);
                int Heading = Convert.ToInt16(Math.Floor(Convert.ToDouble("0" + Parts[8])));

                DeviceMessage deviceMessage = new DeviceMessage(DeviceId,
                    UtcDateTime, Longitude, Latitude, GpsAccuracyInMeters, AltitudeInMeters, SpeedInKph, Heading);

                // Here to get the custom data from the recive string.
                // Eg. CustomDataValue = GetDistance(receiveString);
                string CustomDataValue = "100";
                deviceMessage.CustomData.Add(new CustomData("Distance", CustomDataValue, MimeType.Text));

                string Status = Parts[13].ToString();
                if (Status.ToUpper().StartsWith("SOS"))
                {
                    deviceMessage.Alerts.Add("Panic");
                }
                if (Status.ToUpper().StartsWith("LP"))
                {
                    deviceMessage.Alerts.Add("LowBattery");
                }

                PostResult myResult = mGatewayManager.PostDeviceMessage(deviceMessage);
                Trace.WriteLine(string.Format("The post result code:{0}, message:{1}", myResult.Code.ToString(), myResult.ErrorMessage));
                if (myResult.Code == ResultCode.Error)
                {
                    mGatewayManager.LogDeviceError(DeviceId, "TcpSampleListener", myResult.ErrorMessage);
                }
            }
            catch (Exception Ex)
            {
                Trace.WriteLine(Ex.ToString());
                mGatewayManager.LogDeviceError(DeviceId, "TcpSampleListener", Ex.ToString());
            }
        }

        private void FinallyRecieved(TcpSocketClient Client)
        {
            System.Diagnostics.Trace.WriteLine("Check if client is connected");
            if (Client.Sock.Connected)
            {
                try
                {
                    System.Diagnostics.Trace.WriteLine("Setting up callback started");
                    Client.SetupRecieveCallback(this);
                }
                catch (Exception Ex)
                {
                    System.Diagnostics.Trace.WriteLine("Setting up callback failed");
                    System.Diagnostics.Trace.WriteLine(Ex.ToString());
                    Client.Sock.Close();
                }
                System.Diagnostics.Trace.WriteLine("Setting up callback finished");
            }
            else
            {
                try
                {
                    Client.Sock.Close();
                }
                catch { }

            }

            //RemoveClosedClients();
            System.Diagnostics.Trace.WriteLine("Garbage Collection Start");
            GC.Collect();
            GC.WaitForPendingFinalizers();
            System.Diagnostics.Trace.WriteLine("Garbage Collection Finished");
        }


        /// <summary>
        /// Convert the "Degrees + Minutes + Direction" format to Decimal Degrees
        /// </summary>
        /// <param name="Degrees"></param>
        /// <param name="Minutes"></param>
        /// <param name="Direction"></param>
        /// <returns></returns>
        private decimal DMToDecimalDegrees(double Degrees, double Minutes, string Direction)
        {
            double DecMin;
            double DecDegrees;
            DecMin = Minutes / 60;
            if (Degrees >= 0)
            {
                DecDegrees = Degrees + DecMin;
            }
            else
            {
                DecDegrees = Degrees - DecMin;
            }
            if (Direction == "W" | Direction == "S")
            {
                DecDegrees = DecDegrees * -1;
            }
            return Convert.ToDecimal(DecDegrees);
        }

    }

    /// <summary>
    /// Class holding information and buffers for the Client socket connection
    /// </summary>
    public class TcpSocketClient
    {
        private Socket m_sock;						// Connection to the client
        private byte[] m_byBuff = new byte[64000];		// Receive data buffer
        private string mDeviceId = "";

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="sock">client socket conneciton this object represents</param>
        public TcpSocketClient(Socket sock)
        {
            m_sock = sock;
        }
   
        public string DeviceId
        {
            get { return mDeviceId; }
            set { mDeviceId = value; }
        }

        // Readonly access
        public Socket Sock
        {
            get { return m_sock; }
        }

        /// <summary>
        /// Setup the callback for recieved data and loss of conneciton
        /// </summary>
        /// <param name="app"></param>
        public bool SetupRecieveCallback(TcpSampleListener app)
        {
            try
            {
                AsyncCallback recieveData = new AsyncCallback(app.OnRecievedData);
                m_sock.BeginReceive(m_byBuff, 0, m_byBuff.Length, SocketFlags.None, recieveData, this);
                return true;
            }
            catch (Exception Ex)
            {
                System.Diagnostics.Trace.WriteLine("Recieve callback setup failed! {0}" + Ex.ToString());
                return false;
            }
        }

        /// <summary>
        /// Data has been recieved so we shall put it in an array and
        /// return it.
        /// </summary>
        /// <param name="ar"></param>
        /// <returns>Array of bytes containing the received data</returns>
        public byte[] GetRecievedData(IAsyncResult ar)
        {
            int nBytesRec = 0;
            try
            {
                nBytesRec = m_sock.EndReceive(ar);
            }
            catch (Exception Ex)
            {
                System.Diagnostics.Trace.WriteLine(Ex.ToString());
                return null;
            }
            byte[] byReturn = new byte[nBytesRec];
            Array.Copy(m_byBuff, byReturn, nBytesRec);
            return byReturn;
        }
    }

}
